# +
import itertools as it

import sys
import kwant
import numpy as np
from codes.tb import TB
from codes.tests.test_shortjj import (
    shiba_3D,
    solve_3d_imp,
    solve_3d_selfenergy,
    solve_3d_shortjj,
    solve_exact_3D,
)
from codes.utils import (
    bdg,
    dispersion,
    fourier_func,
    inv_fourier_quad,
    surface_green,
    tau_0,
    tau_x,
    tau_y,
    tau_z,
    taus,
)
from matplotlib import pyplot as plt
from scipy.sparse.linalg import eigsh
from tqdm import tqdm


# -

def Splus(s,ms):
    return (s*(s+1)-ms*(ms+1))**(1/2)
def Smin(s,ms):
    return (s*(s+1)-ms*(ms-1))**(1/2)


def orbitals(l):
    ml =  np.arange(l,-l-1,-1)
    Sz = np.diag(ml)
    Sp = np.diag(Splus(l,ml[1:]),1)
    Sm = np.diag(Smin(l,ml[:-1]),-1)
    Sx = (Sp+Sm)/2
    Sy = -1j*(Sp -Sm)/2
    return Sx, Sy, Sz


# +
def hop_sp(coord, t_sp):
    hop = np.zeros((4,4), dtype='complex')
    hop[0, coord] = t_sp
    hop[coord, 0] = -t_sp
    return -np.kron(tau_0, hop)

def hop_p(coord):
    hop = np.zeros((4,4))
    hop[coord, coord] = 1
    return -np.kron(tau_0, hop)


# -

def shiba_3D_atomistic(
    B=1,
    alpha=0.5,
    onsite_bulk=4 * [0],
    onsite_surf=4 * [2],
    mu=0,
    mu_impurity=0,
    t_barrier=0.75,
    s=0,
    t_sp=0.1,
):
    """
    Creates a simplified n orbital 3D Shiba chain model:
    impurity chain connected only to the superconducting leads.
    The model is returned only in the normal state.

    Parameters:
    -----------
    B: float
        Zeeman splitting in the impurities.
    alpha: float
        spin-orbit coupling strength in the leads.
    mu: list of floats
        chemical potentials px, py, pz in the leads.
    mu_impurity: float
        chemical potential in the impurities.
    t_barrier: float
        hopping between the impurities and the leads.
    hop_imp: float
        hopping between the impurities
    n_orb : int
        amount of orbitals

    Returns:
    --------
    tb_imp: TB
        tight-binding model of the impurity chain.
        Hopping term is between the impurity and the lead.
    tb_bulk: TB
        tight-binding model of the leads.
    """
    # hopping between shiba chain and the lead
    V = np.kron(tau_0, np.array([0, 0, 0, -t_barrier]))

    h_bulk = {}
    s_bulk = {}
    orbs = orbitals(l=1)
    zero_vec = np.zeros(3)
    # bulk hoppings along x and y
    for i in range(3):
        vec = zero_vec.copy()
        vec[i] = 1
        h_bulk[tuple(vec)] = hop_p(i + 1) + hop_sp(i + 1, t_sp)
        h_bulk[tuple(-vec)] = h_bulk[tuple(vec)].T.conj()
        s_bulk[tuple(vec)] = s * np.eye(8)
        s_bulk[tuple(-vec)] = s_bulk[tuple(vec)].T.conj()

    soc_b = (
        np.array(
            [
                [1j / np.sqrt(2), 1/ np.sqrt(2), 0],
                [0, 0, 1],
                [1j / np.sqrt(2), -1/ np.sqrt(2), 0],
            ]
        )
    ).conj().T

    taus = [tau_x, tau_y, tau_z]
    soc = alpha * sum(
        np.kron(taus[i], soc_b @ orbs[i] @ soc_b.conj().T) for i in range(3)
    )

    p_indices = list(range(1, 4)) + list(range(5, 8))
    soc_term = np.zeros((8, 8), dtype="complex")
    soc_term[np.ix_(p_indices, p_indices)] = soc

    global_onsite = -mu * np.eye(8) + soc_term

    h_bulk[tuple(zero_vec)] = np.kron(tau_0, -np.diag(onsite_bulk)) + global_onsite
    s_bulk[tuple(zero_vec)] = np.eye(8)

    h_surf = h_bulk.copy()
    h_surf[tuple(zero_vec)] = np.kron(tau_0, -np.diag(onsite_surf)) + global_onsite

    h_imp = {(0, 0, 0): -mu_impurity * tau_0 + B * tau_z}
    s_imp = {(0, 0, 0): tau_0}
    tb_bulk = TB(h_bulk, s_bulk)
    tb_surf = TB(h_surf, s_bulk)
    tb_imp = TB(h_imp, s_imp)
    tb_V = TB({(0, 0, 1): V.conj().T, (0, 0, -1): V}, {(0, 0, 0): 0 * tau_0})

    return tb_imp, tb_bulk, tb_surf, tb_V


# +
alpha = 0.2
mu_impurity = 0
mu = -1
B = 1
t_sp = 1
t_barrier = 1
epsabs = 1e-4
cutoff = 1e1


def YSR_sigma(k):
    tb_imp, tb_bulk, tb_surf, tb_V = shiba_3D_atomistic(
        alpha=alpha,
        t_sp=t_sp,
        mu_impurity=mu_impurity,
        B=B,
        t_barrier=t_barrier,
        mu=mu,
    )
    return solve_3d_selfenergy(k, tb_bulk, tb_V, tb_surf=tb_surf, x=0, dask_calculation=True, cutoff=cutoff, quad_kwargs={"epsabs": epsabs})


# -

n = 201
k_space = np.linspace(-np.pi, np.pi, n, endpoint=True)
inputs = list(it.product(k_space))

# +
from dask_quantumtinkerer import Cluster

path = "~/scatterer_at_surface/analysis"
cluster = Cluster(101, extra_path=path, asynchronous=False)

cluster.launch_cluster()
# -

client = cluster.get_client()

sys.stdout = open("logfile.txt", "w")

result_ungathered = [client.submit(YSR_sigma, input_tuple[0]) for input_tuple in inputs]

sigma = client.gather(result_ungathered)
from jupyterpost import post
post("ysr done!", "@rikbroekhoven")
np.savetxt(
    f"data/3d_test/atomistic_soc_sigma_alpha{alpha}tsp{t_sp}V{t_barrier}mu{mu}epsabs{epsabs}cutoff{cutoff}",
    np.array(sigma).reshape(
        -1,
    ),
)

print(np.shape(sigma))

# +
font = {"weight": "regular", "size": 14}

plt.rc("font", **font)
plt.plot(k_space, np.imag(np.array(sigma))[:, 0, 0])
plt.xlabel(r'$k_x$')
plt.ylabel(r'$|(\Sigma)|$')
# -

# !ssh hpc05 -C "killall dask-quantumtinkerer-server"
