# +
import matplotlib.pyplot as plt
import numpy as np

from scipy.optimize import curve_fit
from tqdm import tqdm
# -

from chain_decoherence_utils import simple_chain, structure_decay, structure_phase_evolution

# +
B = 0.4  # [T] External field strength

Bangle = 90  # [degrees] Angle of field with respect to MgO lattice
deg = np.pi / 180  # degrees in radians[rad]
Bhat = np.asarray([0, 0, 1])
B = Bhat * B
# -

# ## Single run

nsteps = 1000
N, dist = 4, 5
T = 0.01
times = np.linspace(0, 800, nsteps)
structure = simple_chain(N, dist)
rate_tot, eig, states = structure_decay(structure, B, T, NN_exchange=True)
result = structure_phase_evolution(structure, rate_tot, eig, states, times)

# +
times = np.linspace(0, 800, nsteps)
N = np.shape(structure)[0]
fig, ax = plt.subplots()
colors = ["#5790fc", "#f89c20", "#e42536", "#964a8b", "#9c9ca1", "#7a21dd"]

for i in range(N):
    ax.plot(times, result[i], color=colors[i], label=f"N={i}")
ax.plot(times,(sum(result))-(N-1)*0.5, label='tot')
ax.set_ylabel(r"$\langle S_z \rangle$")
ax.set_xlabel("time (ns)")

plt.annotate(r"$T$ = 0.4 K", (120, 0.4))
#plt.savefig("Images/N4_sym_T_0.4_sym.svg", bbox_inches="tight")
# -

# ## Study of decoherence as function of N and T

# +
def exp(t, A, B, C):
    return C + B * np.exp(-t / A)


def fit_coherence(states, result, tmax=800, nsteps=1000):
    """ evaluate coherence from density matrix and fit with an exponential """
    times = np.linspace(0, tmax, nsteps)
    coherence = np.zeros((nsteps,), dtype="complex")

    for i in range(nsteps):
        coherence[i] = np.sum(np.abs(result[i] - np.diag(np.diag(result[i]))))

    p, _ = curve_fit(exp, times, coherence)
    return p[0]


# +
N_T = 10
N_N = 7
T_min, T_max = 1e-2, 0.2
tmax = 800
nsteps = 1000

t2s_sym = np.zeros((N_N, N_T))
t2s_asym = np.zeros((N_N, N_T))
t2s_no_relax = np.zeros((N_N, N_T))
times = np.linspace(0, tmax, nsteps)

for ind_T, T in tqdm(enumerate(np.linspace(T_min, T_max, N_T))):
    for ind_N, N in enumerate(range(2, 2 + N_N)):

        structure = simple_chain(N, 5)
        rate_tot, eig, states = structure_decay(structure, B, T)
        t2s_sym[ind_N, ind_T] = fit_coherence(
            *structure_phase_evolution(structure, rate_tot, eig, states, times, output_dm=True),
            tmax,
            nsteps
        )
        rate_tot, eig, states = structure_decay(structure, B, T)
        t2s_no_relax[ind_N, ind_T] = fit_coherence(
            *structure_phase_evolution(structure, rate_tot, eig, states, times, output_dm=True, pure_dephasing=False),
            tmax,
            nsteps
        )
# -

np.savetxt(f"Data/Decoherence/t2N8_sym_T{T_min}{T_max}", t2s_sym)
np.savetxt(f"Data/Decoherence/t2N8_no_dephasing_T{T_min}{T_max}", t2s_no_relax)
