import numpy as np
from qutip import Qobj, spin_Jm, spin_Jx, spin_Jz, tensor

from Pulsed_ESR_package.pulsed_esr import (
    h_ident,
    hamiltonian,
    kondo_rates,
    solve_rate_phase_equations,
)

from Redfield_package.SWAPpaperfunction import interaction_strengths, muB, hbar


# +
def stm_chain(N, dist):
    """
    Generate atom positions of a same distance spin chain with a dislocated sensor spin
    
    Parameters:
    N: integer
       amount of spins
    dist: integer 
          number of atomic sites between the spins

    Returns:
    2D array with atom positions of the chain
    """
    atoms = np.linspace(0, dist * (N - 2), N - 1)
    chain = np.zeros((N, 3))
    chain[0] = np.array([-(dist + 0.5), -0.5, 0])
    chain[1:, 0] = atoms
    return chain


def simple_chain(N, dist):
    """
    Generate atom positions of a same distance spin chain
    
    Parameters:
    N: integer
       amount of spins
    dist: integer 
          number of atomic sites between the spins

    Returns:
    2D array with atom positions of the chain
    """
    atoms = np.linspace(0, dist * (N - 1), N)
    chain = np.zeros((N, 3))
    chain[:, 0] = atoms
    return chain


def generate_chain_hamiltonian(structure, B, Btip, N, S, output=False, NN_exchange=True):
    """
    Generate the Hamiltonian of a spin structure
    
    Parameters:
    structure: ndarray
        2D array with atom spin positions
    B : 1D array
        magnetic field
    Btip : 1D array
        tip field
    N: integer
       amount of spins
    S: float 
       spin values
    output: boolean
        if True function prints all coupling strengths
    NN_exchange: boolean
        if True only NN exchange is used for the exchange part of the Hamiltonian    
        
    Returns:
    eig: 1D array
        spin Hamiltonian energies in MHz
    states: 2D array
        eigenstates of spin Hamiltonian
    """
    g = np.array([1.74] * N)
    gyro = (g * muB) / hbar

    J, D, atomcomb = interaction_strengths(structure, gyro, B, N=N, output=output)
    if NN_exchange:
        J = np.array([j * (np.abs(atomcomb[i,0]-atomcomb[i,1])==1) for i, j  in enumerate(J)])
        D = np.array([0.0]*np.shape(atomcomb)[0])
    
    J = np.vstack([(J + 2 * D) / 1e9, (J - D) / 1e9, (J - D) / 1e9]).T
    larmor = np.array(gyro[:, np.newaxis] * B[np.newaxis, :] / (2 * np.pi * 1e9))
    larmor[0, :] += np.array(gyro[0, np.newaxis] * Btip / (2 * np.pi * 1e9))

    ising = False
    H = hamiltonian(larmor, J, N=N, S=S)

    eig, states = H.eigenstates()
    eig /= 2 * np.pi  # convert to frequency
    return eig, states

def structure_decay(
    structure,
    B,
    T=1e-2,
    asymmetric=False,
    Btip=np.array([0, 0, 0]),
    output=False,
    NN_exchange=True,
):
    """
    Generate matrix of transition rates of a spin structure coupled to a spin bath 
    
    Parameters:
    structure: ndarray
        2D array with atom spin positions
    T: float
        temperature
    asymmetric: boolean
        if True first spin has no decoherence
    B: 1D array
        external magnetic field
    Btip: 1D array
        tip field, default value is no tip field
    output: boolean:
        if True function prints all coupling strengths
    NN_exchange: boolean
        if True only NN exchange is used for the exchange part of the Hamiltonian    
        
    Returns:
    rate_tot: 2D array
        array of transition rates due to Kondo spin interaction with the environment
    eig: 1D array
        spin Hamiltonian energies in MHz
    states: 2D array
        eigenstates of spin Hamiltonian
    """

    N = np.shape(structure)[0]
    S = [1 / 2] * N
   
    eig, states = generate_chain_hamiltonian(
        structure=structure, B=B, Btip=Btip, N=N, S=S, output=output, NN_exchange=NN_exchange
    )

    # rates
    rate_tot = 0
    J_surf = [8e-3] * N
    if asymmetric:
        J_surf[0] = 0
    for l in range(N):
        rate_tot += kondo_rates(
            J=[J_surf[l], J_surf[l]],
            V_bias=0,
            eta=[0, 0],
            T=T,
            eig=eig,
            states=states,
            l=l,
            N=N,
            S=S,
        )

    return rate_tot, eig, states


def structure_phase_evolution(
    structure,
    rate_tot,
    eig,
    states,
    times=np.linspace(0, 1e2, 1000),
    pure_dephasing=True,
    output_dm=False,
):
    """
    Generate matrix of transition rates of a spin structure coupled to a spin bath 
    
    Parameters:
    structure: ndarray
        2D array with atom spin positions
    rate_tot: ndarray
        array of transition rates due to Kondo spin interaction with the environment
    eig: 1D array
        spin Hamiltonian energies in MHz
    states: 2D array
        eigenstates of spin Hamiltonian
    times: 1D array
    pure_dephasing: boolean
        if False only inelastic sources of decoherence are considered
    output_dm: boolean
        if True the output is density matrices per timestep

    Returns:
    depending on "output_dm" either the spin z projection or the density matrix is output 
    """
    nsteps = np.shape(times)[0]
    N = np.shape(structure)[0]
    S = [1 / 2] * N
    
    # state initialization
    op_p = h_ident(N, S)
    op_p[0] = spin_Jm(S[0])
    psi_0 = tensor(op_p) @ states[0]
    rho_0 = psi_0 * psi_0.dag()

    rho = solve_rate_phase_equations(
        rate_tot,
        rho_0.transform(states).full(),
        times,
        eig,
        pure_dephasing=pure_dephasing,
    )

    if output_dm:
        return states, rho

    result = np.zeros((N, nsteps), dtype="complex")
    for i in range(N):
        e_op = h_ident(N, S)
        e_op[i] = spin_Jz(S[i])
        e_op = tensor(e_op)
        for j in range(nsteps):
            result[i, j] = np.trace(
                Qobj(rho[j]).transform(states, inverse=True).full() @ e_op.full()
            )
    return result
# -


